/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#include "sockvxd.h"


void
SendCallback (void *Context, TDI_STATUS final, uint ByteCount)
{
	TDI_REQUEST r;
	PSENDCALLBACKMEMO next, p = (PSENDCALLBACKMEMO) Context;
	PSOCKET pSock;
	if (p && p->scbm_socket) {
		/* release old buffers */
		pSock = p->scbm_socket;
		pSock->s_sendstatus = final;
		pSock->s_sendbytes -= p->scbm_len;
		if ((pSock->s_sendbytes < pSock->s_sendlowwater) && pSock->s_post) {
			(*pSock->s_post)++;
			_SignalID ((DWORD) pSock->s_post);
		}

		if (p->scbm_buffer)
			_HeapFree(p->scbm_buffer, 0);
		if (p->scbm_ndis)
			_HeapFree(p->scbm_ndis, 0);
	
		/* if another packet waiting, launch it */
		_asm pushf
		_asm cli
		pSock->s_sendcbm = next = p->scbm_next;
		_asm popf
		if (next) {
			r.Handle.ConnectionContext = pSock->s_conhdl;
			r.RequestNotifyObject = &SendCallback;
			r.RequestContext = (PVOID) next;
			pDT->TdiSendEntry(&r, 0, next->scbm_len, next->scbm_ndis);
		}
		_HeapFree(p, 0);
	}
}

int
sys_send (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	TDI_REQUEST r;
	PSENDCALLBACKMEMO pCBM = _HeapAllocate(sizeof(SENDCALLBACKMEMO), 0);
	PPSENDCALLBACKMEMO backlink = NULL;
	PNDIS_BUFFER pNB = _HeapAllocate(sizeof(NDIS_BUFFER), HEAPZEROINIT);
	TDI_STATUS err;
	DWORD len = pRegs->Client_ECX;
	DWORD pLinBuf = Client_Ptr_Flat(CLIENT_DS, CLIENT_SI);
	PCHAR pBuf = _HeapAllocate(len, 0);

	if (pSock->s_flags & (SOCKFL_BOUND & SOCKFL_CONNECTED) != 
		(SOCKFL_BOUND & SOCKFL_CONNECTED))
		return -SOCKERR_NOT_CONNECTED;

	if (!pNB || !pBuf) 
		return -SOCKERR_NO_MEMORY;

	if (!pLinBuf)
		return -SOCKERR_ACCESS;

	if (pSock->s_sendbytes + len > pSock->s_sendmax) {
		if (pSock->s_flags & SOCKFL_NONBLOCK)
			return -SOCKERR_WOULD_BLOCK;
		do {
			wait (pSock, POLL_WRITE);
		} while (pSock->s_sendbytes + len > pSock->s_sendmax);
		/* XXX--note, above loop can fail to terminate, ever, if
			len is bigger than sendmax */
	}
	pSock->s_sendbytes += len;

	memcpy (pBuf, (PCHAR) pLinBuf, len);

	pNB->VirtualAddress = (PVOID) pBuf;
	pNB->Length = len;

	pCBM->scbm_next = NULL;
	pCBM->scbm_socket = pSock;
	pCBM->scbm_buffer = pBuf;
	pCBM->scbm_len = len;
	pCBM->scbm_ndis = pNB;

	backlink = &pSock->s_sendcbm;
	_asm pushf
	_asm cli
	while (*backlink)
		backlink = &((*backlink)->scbm_next);
	*backlink = pCBM;
	_asm popf

	pRegs->Client_ECX = (pSock->s_sendstatus << 16) | pSock->s_sendbytes;

	/* if first in the queue, launch */
	if (backlink == &pSock->s_sendcbm) {
		r.Handle.ConnectionContext = pSock->s_conhdl;
		r.RequestNotifyObject = &SendCallback;
		r.RequestContext = (PVOID) pCBM;
		err = pDT->TdiSendEntry (&r, 0, len, pNB);
		if ((err != TDI_SUCCESS) && (err != TDI_PENDING)) {
			_HeapFree(pCBM, 0);
			return -err;
		}
	}

	return len;
}


int
sys_receive(HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	PCHAR pLinBuf = (PCHAR) Client_Ptr_Flat(CLIENT_DS, CLIENT_SI);
	DWORD len;

	if (!pLinBuf)
		return -SOCKERR_ACCESS;

	if (pSock->s_flags & SOCKFL_NONBLOCK) {
		len = GetCBuf (pSock->s_recvbuf, (PCHAR) pLinBuf, 
			pRegs->Client_ECX);
		if (len)
			return pRegs->Client_ECX = len;
		if (pSock->s_flags & SOCKFL_PEERCLOSED)
			return pRegs->Client_ECX = 0;
		return -SOCKERR_WOULD_BLOCK;
	}

	do {
		wait (pSock, POLL_READ);
		len = GetCBuf (pSock->s_recvbuf, (PCHAR) pLinBuf, 
			pRegs->Client_ECX);
	} while ((len== 0) && !(pSock->s_flags & SOCKFL_PEERCLOSED));

	return pRegs->Client_ECX = len;
}

TDI_STATUS Receive(PVOID AddrContext, PVOID ConnectionContext, ulong Flags, 
				   uint Indicated, uint Available, uint *Taken, uchar *Data, 
				   EventRcvBuffer *Buffer)
{
	PSOCKET pSock = (PSOCKET) ConnectionContext;
	DWORD len;

	if (!pSock || pSock->s_signature != SOCKSIG)
		return TDI_NOT_ACCEPTED;

	if (!pSock->s_recvbuf)
		return TDI_NOT_ACCEPTED;

	if (!TryLockCBuf (pSock->s_recvbuf))
		return TDI_NOT_ACCEPTED;

	len = Indicated;
	if ((int) len >= AvailCBuf (pSock->s_recvbuf))
		len = AvailCBuf (pSock->s_recvbuf) - 1;

	PutCBuf (pSock->s_recvbuf, Data, len);
	*Taken = len;
	UnlockCBuf (pSock->s_recvbuf);

	if (pSock->s_post) {
		(*pSock->s_post)++;
		_SignalID((DWORD) pSock->s_post);
	}

	return TDI_SUCCESS;
}

